@@ -274,7 +274,7 @@ int parse_init(int argc, char **argv, void **context)
 	if ( !noptstr )
 	  break;
 	noptstr[optlen] = ',';
-	strcpy(noptstr+len+1, argv[i]);
+	strcpy(noptstr+optlen+1, argv[i]);
 	optlen += len+1;
       } else {
 	noptstr = (char *) malloc(len+1);
@@ -291,8 +291,8 @@ int parse_init(int argc, char **argv, void **context)
     }
   }
 
-  /* We only need this once */
-
+  /* We only need this once.  NFS mounts are so common that we cache
+     this module. */
   if ( !mount_nfs )
     if ( (mount_nfs = open_mount("nfs", MODPREFIX)) ) {
       init_ctr++;
@@ -308,22 +308,26 @@ int parse_init(int argc, char **argv, void **context)
 int parse_mount(char *root, char *name, int name_len, char *mapent, void *context)
 {
   struct parse_context *ctxt = (struct parse_context *) context;
-  char *pmapent, *options, *noptions, *p;
+  char *pmapent, *options, *noptions, *ent, *p, *q, *fstype;
   int mapent_len, rv;
   int l, optlen;
 
   mapent_len = expandsunent(mapent,NULL,name,ctxt->subst);
-  pmapent = malloc(mapent_len + 1);
+  pmapent = alloca(mapent_len + 1);
   if (!pmapent) {
-    syslog(LOG_ERR, MODPREFIX "malloc: %m");
+    syslog(LOG_ERR, MODPREFIX "alloca: %m");
     return 1;
   }
   expandsunent(mapent,pmapent,name,ctxt->subst);
 
   syslog(LOG_DEBUG, "expanded entry: %s", pmapent);
 
-  options = ctxt->optstr;	/* No options found */
-  optlen = options ? strlen(options) : 0;
+  options = strdup(ctxt->optstr ? ctxt->optstr : "");
+  if ( !options ) {
+    syslog(LOG_ERR, MODPREFIX "strdup: %m");
+    return 1;
+  }
+  optlen = strlen(options);
 
   p = pmapent;
   while ( (p = skipspace(p)) ) {
@@ -331,57 +335,90 @@ int parse_mount(char *root, char *name, int name_len, char *mapent, void *contex
       l = chunklen(++p);
       if ( !l )
 	continue;
-      if ( options ) {
-	if ( options != ctxt->optstr ) {
-	  noptions = realloc(options, optlen+l+2);
-	  if ( !noptions ) {
-	    free(pmapent);
-	    free(options);
-	    syslog(LOG_ERR, MODPREFIX "realloc: %m");
-	    return 1;
-	  }
-	} else {
-	  noptions = malloc(optlen+l+2);
-	  if ( !noptions ) {
-	    free(pmapent);
-	    syslog(LOG_ERR, MODPREFIX "malloc: %m");
-	    return 1;
-	  }	  
-	  memcpy(noptions, ctxt->optstr, optlen+1);
-	}
-	noptions[optlen++] = ',';
-	memcpy(noptions+optlen, p, l);
-	noptions[optlen += l] = '\0';
-	options = noptions;
-      } else {
-	options = malloc(l+1);
-	if ( !options ) {
-	  free(pmapent);
-	  syslog(LOG_ERR, MODPREFIX "malloc: %m");
-	  return 1;
-	}
-	memcpy(options, p, l);
-	options[optlen = l] = '\0';
+      noptions = realloc(options, optlen+l+(optlen ? 2 : 1));
+      if ( !noptions ) {
+	free(options);
+	syslog(LOG_ERR, MODPREFIX "realloc: %m");
+	return 1;
       }
+      if ( optlen ) noptions[optlen++] = ',';
+      memcpy(noptions+optlen, p, l);
+      noptions[optlen += l] = '\0';
+      options = noptions;
       p += l;
       syslog(LOG_DEBUG, MODPREFIX "gathered options: %s", options);
     } else if ( *p == '/' ) {
       /* Uh-oh, a multientry; don't support now, but ****FIXME**** */
-      free(pmapent);
-      if ( options != ctxt->optstr ) free(options);
-      syslog(LOG_NOTICE, MODPREFIX "entry %s is a multipath entry", name);
-      return 1;
+      /* We do tolerate "multientries" that only contain / though... */
+      if ( chunklen(p) == 1 ) {
+	p++;
+      } else {
+	free(options);
+	syslog(LOG_NOTICE, MODPREFIX "entry %s is a multipath entry", name);
+	return 1;
+      }
     } else {
       p[chunklen(p)] = '\0';	/* We don't use anything after the entry */
-      syslog(LOG_DEBUG, MODPREFIX "core of entry: %s", p);
-      break;			/* We have the beef of the entry, run w/it */
+      if ( *p == ':' ) p++;	/* Sun escape for entries starting with / */
+      if ( ! *p ) {
+	syslog(LOG_ERR, MODPREFIX "entry %s is empty!", name);
+	free(options);
+	return 1;
+      }
+      ent = p;
+      syslog(LOG_DEBUG, MODPREFIX "core of entry: %s", ent);
+      break;			/* We have the core entry, run w/it */
     }
   }
 
-  rv = mount_nfs->mount_mount(root, name, name_len, p, "nfs", options,
-			      mount_nfs->context);
-  free(pmapent);
-  if ( options != ctxt->optstr ) free(options);
+  fstype = "nfs";		/* Default filesystem type */
+
+  /* Scan for -fstype= option, which isn't a real mount option */
+
+  noptions = strdup(options);
+  if ( !noptions ) {
+    syslog(LOG_ERR, MODPREFIX "strdup: %m");
+    free(options);
+    return 1;
+  }
+
+  p = options;
+
+  for ( q = strtok(noptions, ",") ; q ; q = strtok(NULL, ",") ) {
+    if ( !strncmp(q, "fstype=", 7) ) {
+      fstype = alloca(strlen(q+7)+1);
+      if ( !fstype ) {
+	syslog(LOG_ERR, MODPREFIX "alloca: %s");
+	return 1;
+      }
+      strcpy(fstype, q+7);
+    } else {
+      l = strlen(q);
+      memcpy(p, q, l);
+      p += l;
+      *p = ',';
+      p++;
+    }
+  }
+
+  if ( p > options ) {
+    --p;			/* Delete last comma */
+    *p = '\0';
+  } else {
+    free(options);		/* No options - mount modules want NULL here */
+    options = NULL;
+  }
+  free(noptions);
+
+  if ( !strcmp(fstype, "nfs") ) {
+    rv = mount_nfs->mount_mount(root, name, name_len, ent, fstype, options,
+				mount_nfs->context);
+  } else {
+    /* Generic mount routine */
+    rv = do_mount(root, name, name_len, ent, fstype, options);
+  }
+
+  if (options) free(options);
   return rv;
 }
 
